from __future__ import annotations

import json
import logging
from collections.abc import Callable
from inspect import getfullargspec
from typing import Any

from channels_rpc import logs
from channels_rpc.exceptions import (
    GENERIC_APPLICATION_ERROR,
    INVALID_REQUEST,
    RPC_ERRORS,
    JsonRpcError,
    generate_error_response,
)
from channels_rpc.rpc_base import RpcBase
from channels_rpc.utils import create_json_rpc_frame

logger = logging.getLogger("django.channels.rpc")


class AsyncRpcBase(RpcBase):
    async def execute_called_method(self, method: Callable, params: dict | list) -> Any:
        func_args = getfullargspec(method).varkw
        if func_args and "kwargs" in func_args:
            return (
                await method(*params, consumer=self)
                if isinstance(params, list)
                else await method(**params, consumer=self)
            )
        elif isinstance(params, list):
            return await method(*params)
        else:
            return await method(**params)

    async def process_call(
        self, data: dict[str, Any], *, is_notification: bool = False
    ):
        method = self.get_method(data, is_notification=is_notification)
        params = self.get_params(data)
        rpc_id, rpc_id_key = self.get_rpc_id(data)
        logger.debug(f"Executing {method.__qualname__}({json.dumps(params)})")
        result = await self.execute_called_method(method, params)
        if not is_notification:
            logger.debug("Execution result: %s", result)
            result = create_json_rpc_frame(
                result=result,
                rpc_id=rpc_id,
                rpc_id_key=rpc_id_key,
                method=data["method"],
                params=params,
            )
        elif result is not None:
            logger.warning("The notification method shouldn't return any result")
            logger.warning(f"method: {method.__qualname__}, params: {params}")
            result = None
        return result

    async def intercept_call(
        self, data: dict[str, Any] | list[str, Any] | None
    ) -> tuple[Any, bool]:
        logger.debug("Intercepting call: %s", data)
        if not data:
            logger.warning(logs.EMPTY_CALL)
            message = RPC_ERRORS[INVALID_REQUEST]
            result = generate_error_response(
                rpc_id=None, code=INVALID_REQUEST, message=message
            )
            return result, False
        response = None
        if isinstance(data, dict) and "response" in data:
            response = data["response"]
            if response is not None:
                logger.debug(f"Received RPC response: {response}")
                return response, True
        if isinstance(data, dict) and "request" in data:
            request = data["request"]
            if request is not None:
                logger.debug(f"Received RPC request: {request}")
        result = None
        is_notification: bool = False
        rpc_id = request.get("id") or request.get("call_id")
        method_name = request.get("method")
        logger.debug(logs.CALL_INTERCEPTED, request)
        if isinstance(data, dict):
            is_notification = method_name is not None and rpc_id is None
            if rpc_id:
                logger.info(logs.RPC_METHOD_CALL_START, method_name, rpc_id)
            else:
                logger.info(logs.RPC_NOTIFICATION_START, method_name)
            try:
                result = await self.process_call(
                    request, is_notification=is_notification
                )
            except JsonRpcError as e:
                result = e.as_dict()
            except Exception as e:
                logger.debug("Application error: %s", e)
                exception_data = e.args[0] if len(e.args) == 1 else e.args
                result = generate_error_response(
                    rpc_id=rpc_id,
                    code=GENERIC_APPLICATION_ERROR,
                    message=str(e),
                    data=exception_data,
                )
        # elif isinstance(data, list):
        #     # TODO: implement batch calls
        #     invalid_calls = [x for x in data if not isinstance(x, dict)]
        #     if invalid_calls:
        #         message = RPC_ERRORS[INVALID_REQUEST]
        #         result = generate_error_response(
        #             rpc_id=None, code=INVALID_REQUEST, message=message
        #         )
        if rpc_id:
            logger.debug(logs.RPC_METHOD_CALL_END, rpc_id, method_name, result)
        else:
            logger.debug(logs.RPC_NOTIFICATION_END, method_name)
        return result, is_notification

    async def _base_receive_json(self, data: dict[str, Any]) -> None:
        logger.debug("Received JSON: %s", data)
        result, is_notification = await self.intercept_call(data)
        if not is_notification:
            logger.debug("Sending result: %s", result)
            await self.send_json(result)



import json

from channels.generic.http import AsyncHttpConsumer

from channels_rpc.exceptions import (
    INVALID_REQUEST,
    PARSE_ERROR,
    RPC_ERRORS,
    generate_error_response,
)
from channels_rpc.rpc_base import RpcBase


class AsyncRpcHttpConsumer(AsyncHttpConsumer, RpcBase):
    async def handle(self, body):
        """
        Called on HTTP request
        :param message: message received
        :return:
        """
        if body:
            try:
                data = json.loads(body)
            except ValueError:
                # json could not decoded
                result = generate_error_response(
                    None, PARSE_ERROR, RPC_ERRORS[PARSE_ERROR]
                )
            else:
                result, is_notification = self.intercept_call(data)

            # Set response status code
            # http://www.jsonrpc.org/historical/json-rpc-over-http.html#response-codes
            if is_notification:
                # notification response
                status_code = 204
                if result and "error" in result:
                    status_code = self.RPC_ERROR_TO_HTTP_CODE[result["error"]["code"]]
                result = ""
            elif "error" in result:
                status_code = self.RPC_ERROR_TO_HTTP_CODE[result["error"]["code"]]
            else:
                status_code = 200
        else:
            result = generate_error_response(
                None, INVALID_REQUEST, RPC_ERRORS[INVALID_REQUEST]
            )

        self.send_response(
            status_code,
            json.dumps(result),
            headers=[
                (b"Content-Type", b"application/json-rpc"),
            ],
        )



EMPTY_CALL: str = "Received empty remote procedure call data."
CALL_INTERCEPTED: str = "Received RPC: %s"
RPC_METHOD_CALL_START: str = "Calling the '%s' method with RPC ID #%s."
RPC_METHOD_CALL_END: str = "RPC ID #%s for method '%s' processed with result: %s."
RPC_NOTIFICATION_START: str = "Notification method '%s' received."
RPC_NOTIFICATION_END: str = "Notification method '%s' processed."
INVALID_JSON_RPC_VERSION: str = "Invalid JSON-RPC version! Expected '2.0', got '%s'."

# flakes8: noqa: E501



"""Exceptions for the channels-rpc package."""
import json
from typing import Any

from channels_rpc.utils import create_json_rpc_frame

PARSE_ERROR: int = -32700
INVALID_REQUEST: int = -32600
METHOD_NOT_FOUND: int = -32601
INVALID_PARAMS: int = -32602
INTERNAL_ERROR: int = -32603
GENERIC_APPLICATION_ERROR: int = -32000
PARSE_RESULT_ERROR: int = -32701
RPC_ERRORS: dict[int, str] = {
    PARSE_ERROR: "Parse Error",
    INVALID_REQUEST: "Invalid Request",
    METHOD_NOT_FOUND: "Method Not Found",
    INVALID_PARAMS: "Invalid Params",
    INTERNAL_ERROR: "Internal Error",
    GENERIC_APPLICATION_ERROR: "Application Error",
    PARSE_RESULT_ERROR: "Error while parsing result",
}


def generate_error_response(
    rpc_id: int, code: int, message: str, data=None
) -> dict[str, Any]:
    """Generate a JSON-RPC error response.

    Parameters
    ----------
    rpc_id : int
        Call ID.
    code : int
        RPC error code.
    message : str
        Error message.
    data : Any, optional
        Additional error data, by default None.

    Returns
    -------
    dict[str, Any]
        Error response.
    """
    error = {"code": code, "message": message}
    if data is not None:
        error["data"] = data
    return create_json_rpc_frame(error=error, rpc_id=rpc_id)


class JsonRpcError(Exception):
    """General JSON-RPC exception class."""

    def __init__(self, rpc_id: int, code: int):
        """Initialize a new :class:`JsonRpcError` instance.

        Parameters
        ----------
        rpc_id : int
            Call ID.
        code : int
            RPC error code.
        data : _type_, optional
            _description_, by default None
        """
        self.rpc_id = rpc_id
        self.code = code

    def as_dict(self) -> dict[str, Any]:
        """Return an error response dictionary.

        Returns
        -------
        dict[str, Any]
            Error response.
        """
        message = RPC_ERRORS[self.code]
        return generate_error_response(
            rpc_id=self.rpc_id, code=self.code, message=message
        )

    def __str__(self) -> str:
        """Error response dictionary as a string.

        Returns
        -------
        str
            Error response.
        """
        return json.dumps(self.as_dict())



"""Definition of the :class:`RpcBase` class.

A JSON-RPC request message can contain three possible elements: The *method*, which
is a string that names the method to be invoked; *params*, which are objects or arrays
of values that get passed along as parameters to the destination app; and *id*, a
string or number that matches the response with the request that it is replying to.

References
----------
- https://nonamesecurity.com/learn/what-is-json-rpc/
"""
from __future__ import annotations

import json
import logging
from collections import defaultdict
from collections.abc import Callable
from inspect import getfullargspec
from typing import Any

from six import string_types

from channels_rpc import logs
from channels_rpc.exceptions import (
    GENERIC_APPLICATION_ERROR,
    INTERNAL_ERROR,
    INVALID_PARAMS,
    INVALID_REQUEST,
    METHOD_NOT_FOUND,
    PARSE_ERROR,
    RPC_ERRORS,
    JsonRpcError,
    generate_error_response,
)
from channels_rpc.utils import create_json_rpc_frame

logger = logging.getLogger("django.channels.rpc")


class RpcBase:
    """Base class for RPC consumers.

    Variant of WebsocketConsumer that automatically JSON-encodes and decodes
    messages as they come in and go out. Expects everything to be text; will
    error on binary data.

    Errors:

    - -32700
      Parse error
      Invalid JSON was received by the server.
      An error occurred on the server while parsing the JSON text.

    - -32600
      Invalid Request
      The JSON sent is not a valid Request object.

    - -32601
      Method not found
      The method does not exist / is not available.

    - -32602
      Invalid params
      Invalid method parameter(s).

    - -32603
      Internal error
      Internal JSON-RPC error.

    - -32099 to -32000
      Server error
      Reserved for implementation-defined server-errors. (@TODO)

    References
    ----------
    - http://groups.google.com/group/json-rpc/web/json-rpc-2-0
    """

    RPC_ERROR_TO_HTTP_CODE: dict[int, int] = {
        PARSE_ERROR: 500,
        INVALID_REQUEST: 400,
        METHOD_NOT_FOUND: 404,
        INVALID_PARAMS: 500,
        INTERNAL_ERROR: 500,
        GENERIC_APPLICATION_ERROR: 500,
    }
    RPC_ID_KEYS: list[str] = ["id", "call_id"]

    rpc_methods: defaultdict = defaultdict(dict)
    rpc_notifications: defaultdict = defaultdict(dict)

    @classmethod
    def rpc_method(
        cls,
        method_name: str | None = None,
        *,
        websocket: bool = True,
        http: bool = True,
    ) -> Callable:
        """A decorator for registering RPC methods.

        Parameters
        ----------
        method_name : str, optional
            RPC method name for the function, by default None.
        websocket : bool, optional
            Whether WebSocket transport can use this function, by default True.
        http : bool, optional
            Whether HTTP transport can use this function, by default True.

        Returns
        -------
        Callable
            Decorated function.
        """

        def wrap(method: Callable) -> Callable:
            name = method_name or method.__name__
            method.options = {"websocket": websocket, "http": http}
            cls.rpc_methods[id(cls)][name] = method
            return method

        return wrap

    @classmethod
    def get_rpc_methods(cls) -> list[str]:
        """List RPC methods available for this consumer.

        Returns
        -------
        list[str]
            List of RPC methods available for this consumer.
        """
        try:
            return list(cls.rpc_methods[id(cls)])
        except KeyError:
            return []

    @classmethod
    def rpc_notification(
        cls,
        notification_name: str | None = None,
        *,
        websocket: bool = True,
        http: bool = True,
    ) -> Callable:
        """A decorator for registering RPC notifications.

        Parameters
        ----------
        notification_name : str, optional
            RPC name for the function, by default None.
        websocket : bool, optional
            Whether WebSocket transport can use this function, by default True.
        http : bool, optional
            Whether HTTP transport can use this function, by default True.

        Returns
        -------
        Callable
            Decorated function.
        """

        def wrap(method: Callable) -> Callable:
            name = notification_name or method.__name__
            method.options = {"websocket": websocket, "http": http}
            cls.rpc_notifications[id(cls)][name] = method
            return method

        return wrap

    @classmethod
    def get_rpc_notifications(cls) -> list[str]:
        """List RPC notifications available for this consumer.

        Returns
        -------
        list[str]
            List of RPC notifications available for this consumer.
        """
        try:
            return list(cls.rpc_notifications[id(cls)])
        except KeyError:
            return []

    def notify_channel(self, method: str, params: dict[str, Any]) -> None:
        """Notify a channel.

        Parameters
        ----------
        method : str
            Method name.
        params : dict[str, Any]
            Method parameters.
        """
        content = create_json_rpc_frame(method=method, params=params)
        self.send(self.encode_json(content))

    def validate_call(self, data: dict[str, Any]) -> None:
        """Validate RPC call data.

        Parameters
        ----------
        data : dict[str, Any]
            Remote procedure call data.

        Raises
        ------
        JsonRpcError
            Invalid call data.
        """
        logger.debug("Validating call data: %s", data)
        rpc_id = data.get("id")
        bad_json_rpc_version = data.get("jsonrpc") != "2.0"
        no_method = "method" not in data
        bad_method = not isinstance(data.get("method"), string_types)
        if bad_json_rpc_version:
            logger.warning(logs.INVALID_JSON_RPC_VERSION, data.get("jsonrpc"))
        if no_method or bad_method:
            raise JsonRpcError(rpc_id, INVALID_REQUEST)
        logger.debug("Call data is valid")

    def get_method(self, data: dict[str, Any], *, is_notification: bool) -> Callable:
        """Get the method to call.

        Parameters
        ----------
        data : dict[str, Any]
            Remote procedure call data.
        is_notification : bool
            Whether the call is a notification.

        Returns
        -------
        Callable
            Method to call.

        Raises
        ------
        JsonRpcError
            Invalid call data provided.
        JsonRpcError
            RPC method not supported.
        """
        self.validate_call(data)
        rpc_id = data.get("id") or data.get("call_id")
        method_name = data["method"]
        logger.debug("Getting method: %s", method_name)
        class_id = id(self.__class__)
        methods = self.rpc_notifications if is_notification else self.rpc_methods
        try:
            method = methods[class_id][method_name]
        except KeyError as e:
            raise JsonRpcError(rpc_id, METHOD_NOT_FOUND) from e
        protocol = self.scope["type"]
        if not method.options[protocol]:
            raise JsonRpcError(rpc_id, METHOD_NOT_FOUND)
        logger.debug("Method found: %s", method.__name__)
        return method

    def get_params(self, data: dict[str, Any]) -> dict | list:
        """Get the parameters to pass to the method.

        Parameters
        ----------
        data : dict[str, Any]
            Remote procedure call data.

        Returns
        -------
        dict | list
            Parameters to pass to the method.

        Raises
        ------
        JsonRpcError
            Invalid call data provided.
        """
        logger.debug("Getting call parameters: %s", data)
        params = data.get("params") or data.get("arguments") or {}
        if not isinstance(params, (list, dict)):
            rpc_id = data.get("id")
            raise JsonRpcError(rpc_id, INVALID_PARAMS)
        logger.debug("Call parameters found: %s", params)
        return params

    def get_rpc_id(self, data: dict[str, Any]) -> tuple[str | None, str | None]:
        """Get the RPC ID.

        Parameters
        ----------
        data : dict[str, Any]
            Remote procedure call data.

        Returns
        -------
        tuple[str | None, str | None]
            RPC ID and RPC ID key.
        """
        logger.debug("Extracting RPC ID: %s", data)
        for rpc_id_key in self.RPC_ID_KEYS:
            if rpc_id_key in data:
                logger.debug(
                    "RPC ID with key '%s' found: %s", rpc_id_key, data[rpc_id_key]
                )
                return data[rpc_id_key], rpc_id_key

    def process_call(
        self, data: dict[str, Any], *, is_notification: bool = False
    ) -> dict | None:
        """Process the received remote procedure call data.

        Parameters
        ----------
        data : dict[str, Any]
            Remote procedure call data.
        is_notification : bool, optional
            Whether the call is a notification, by default False.

        Returns
        -------
        dict | None
            Result of the remote procedure call.
        """
        method = self.get_method(data, is_notification=is_notification)
        params = self.get_params(data)
        rpc_id, rpc_id_key = self.get_rpc_id(data)
        logger.debug(f"Executing {method.__qualname__}({json.dumps(params)})")
        result = self.execute_called_method(method, params)
        if not is_notification:
            logger.debug("Execution result: %s", result)
            result = create_json_rpc_frame(
                result=result,
                rpc_id=rpc_id,
                rpc_id_key=rpc_id_key,
                method=data["method"],
                params=params,
            )
        elif result is not None:
            logger.warning("The notification method shouldn't return any result")
            logger.warning(f"method: {method.__qualname__}, params: {params}")
            result = None
        return result

    def intercept_call(self, data: dict[str, Any]) -> tuple[Any, bool]:
        """Handles calls and notifications.

        Parameters
        ----------
        data : dict[str, Any]
            Received message data.

        Returns
        -------
        tuple[Any, bool]
            Result of the remote procedure call and whether it is a notification.
        """
        logger.debug("Intercepting call: %s", data)
        if not data:
            logger.warning(logs.EMPTY_CALL)
            message = RPC_ERRORS[INVALID_REQUEST]
            result = generate_error_response(
                rpc_id=None, code=INVALID_REQUEST, message=message
            )
            return result, False
        response = None
        if isinstance(data, dict) and "response" in data:
            response = data["response"]
            if response is not None:
                logger.debug(f"Received RPC response: {response}")
                return response, True
        if isinstance(data, dict) and "request" in data:
            request = data["request"]
            if request is not None:
                logger.debug(f"Received RPC request: {request}")
        result = None
        is_notification: bool = None
        rpc_id = request.get("id") or request.get("call_id")
        method_name = request.get("method")
        logger.debug(logs.CALL_INTERCEPTED, request)
        if isinstance(request, dict):
            is_notification = method_name is not None and rpc_id is None
            if rpc_id:
                logger.info(logs.RPC_METHOD_CALL_START, method_name, rpc_id)
            else:
                logger.info(logs.RPC_NOTIFICATION_START, method_name)
            try:
                result = self.process_call(request, is_notification=is_notification)
            except JsonRpcError as e:
                result = e.as_dict()
            except Exception as e:
                logger.debug("Application error: %s", e)
                exception_data = e.args[0] if len(e.args) == 1 else e.args
                result = generate_error_response(
                    rpc_id=rpc_id,
                    code=GENERIC_APPLICATION_ERROR,
                    message=str(e),
                    data=exception_data,
                )
        # elif isinstance(data, list):
        #     # TODO: implement batch calls
        #     invalid_call_data = [x for x in data if not isinstance(x, dict)]
        #     if invalid_call_data:
        #         message = RPC_ERRORS[INVALID_REQUEST]
        #         result = generate_error_response(
        #             rpc_id=None, code=INVALID_REQUEST, message=message
        #         )
        if rpc_id:
            logger.debug(logs.RPC_METHOD_CALL_END, rpc_id, method_name, result)
        else:
            logger.debug(logs.RPC_NOTIFICATION_END, method_name)
        return result, is_notification

    def execute_called_method(self, method: Callable, params: list | dict) -> Any:
        """Get the result of the remote procedure call.

        Parameters
        ----------
        method : Callable
            Method to call.
        params : list | dict
            Parameters to pass to the method.

        Returns
        -------
        Any
            Result of the remote procedure call.
        """
        func_args = getfullargspec(method).varkw
        if func_args and "kwargs" in func_args:
            return (
                method(*params, consumer=self)
                if isinstance(params, list)
                else method(**params, consumer=self)
            )
        elif isinstance(params, list):
            return method(*params)
        else:
            return method(**params)

    def _base_receive_json(self, data: dict[str, Any]) -> None:
        """Called when receiving a message.

        Parameters
        ----------
        data : dict[str, Any]
            Received message data.
        """
        logger.debug("Received JSON message: %s", data)
        result, is_notification = self.intercept_call(data)
        if not is_notification:
            logger.debug("Sending result: %s", result)
            self.send_json(result)



from channels.generic.websocket import AsyncJsonWebsocketConsumer

from channels_rpc.async_rpc_base import AsyncRpcBase


class AsyncJsonRpcWebsocketConsumer(AsyncJsonWebsocketConsumer, AsyncRpcBase):
    async def receive_json(self, content):
        await self._base_receive_json(content)



from __future__ import annotations

import json
from typing import Any

from channels.generic.websocket import JsonWebsocketConsumer

from channels_rpc.exceptions import (
    PARSE_ERROR,
    PARSE_RESULT_ERROR,
    RPC_ERRORS,
    generate_error_response,
)
from channels_rpc.rpc_base import RpcBase


class JsonRpcWebsocketConsumer(JsonWebsocketConsumer, RpcBase):
    def decode_json(self, data: str | bytes | bytearray) -> Any:
        """Decode remote procedure call data.

        Parameters
        ----------
        data : str | bytes | bytearray
            Remote procedure call data.

        Returns
        -------
        Any
            Decoded remote procedure call data.
        """
        try:
            return json.loads(data)
        except json.decoder.JSONDecodeError:
            frame = generate_error_response(None, PARSE_ERROR, RPC_ERRORS[PARSE_ERROR])
            self.send_json(frame)

    def encode_json(self, data: dict[str, Any]) -> str:
        """Encode remote procedure call data.

        Parameters
        ----------
        data : dict[str, Any]
            Remote procedure call data.

        Returns
        -------
        str
            Encoded remote procedure call data.
        """
        try:
            return json.dumps(data)
        except TypeError:
            frame = generate_error_response(
                None,
                PARSE_ERROR,
                RPC_ERRORS[PARSE_RESULT_ERROR],
                str(data["result"]),
            )
            return json.dumps(frame)

    def receive_json(self, data: dict[str, Any]) -> None:
        self._base_receive_json(data)



from channels_rpc.async_json_rpc_websocket_consumer import AsyncJsonRpcWebsocketConsumer
from channels_rpc.async_rpc_http_consumer import AsyncRpcHttpConsumer
from channels_rpc.exceptions import JsonRpcError
from channels_rpc.json_rpc_websocket_consumer import JsonRpcWebsocketConsumer

__all__ = [
    "AsyncJsonRpcWebsocketConsumer",
    "AsyncRpcHttpConsumer",
    "JsonRpcError",
    "JsonRpcWebsocketConsumer",
]



from __future__ import annotations

from typing import Any

FRAME_BASE: dict[str, Any] = {"jsonrpc": "2.0"}


def create_json_rpc_frame(
    rpc_id: int | None = None,
    result: Any = None,
    params: dict[str, Any] | None = None,
    method: str | None = None,
    error: dict[str, int | str] | None = None,
    rpc_id_key: str = "call_id",
) -> dict[str, Any]:
    frame = FRAME_BASE.copy()
    if result is None:
        return {
            "request": {
                rpc_id_key: rpc_id,
                "method": method,
                "arguments": params or {},
            },
            **frame,
        }
    else:
        return {
            "request": {
                rpc_id_key: rpc_id,
                "method": method,
                "arguments": params or {},
            },
            "response": {
                "result": result or error,
                "result_type": type(result).__name__ if result else None,
                rpc_id_key: rpc_id,
            },
            **frame,
        }
